@@ -1,44 +1,44 @@
-/* $PostgreSQL: pgsql/src/interfaces/ecpg/ecpglib/prepare.c,v 1.26.2.2 2008/06/03 20:55:41 tgl Exp $ */
+/* src/interfaces/ecpg/ecpglib/prepare.c */
 
 #define POSTGRES_ECPG_INTERNAL
 #include "postgres_fe.h"
 
 #include <ctype.h>
 
-#include "ecpgtype.h"
-#include "ecpglib.h"
 #include "ecpgerrno.h"
-#include "extern.h"
+#include "ecpglib.h"
+#include "ecpglib_extern.h"
+#include "ecpgtype.h"
 #include "sqlca.h"
 
-struct prepared_statement
-{
-	char	   *name;
-	bool		prepared;
-	struct statement *stmt;
-	struct prepared_statement *next;
-};
-
 #define STMTID_SIZE 32
 
+/*
+ * The statement cache contains stmtCacheNBuckets hash buckets, each
+ * having stmtCacheEntPerBucket entries, which we recycle as needed,
+ * giving up the least-executed entry in the bucket.
+ * stmtCacheEntries[0] is never used, so that zero can be a "not found"
+ * indicator.
+ */
+#define stmtCacheNBuckets		2039	/* should be a prime number */
+#define stmtCacheEntPerBucket	8
+
+#define stmtCacheArraySize (stmtCacheNBuckets * stmtCacheEntPerBucket + 1)
+
 typedef struct
 {
 	int			lineno;
 	char		stmtID[STMTID_SIZE];
 	char	   *ecpgQuery;
-	long		execs;			/* # of executions		*/
-	char	   *connection;		/* connection for the statement		*/
+	long		execs;			/* # of executions */
+	const char *connection;		/* connection for the statement */
 } stmtCacheEntry;
 
 static int	nextStmtID = 1;
-static const int stmtCacheNBuckets = 2039;		/* # buckets - a prime # */
-static const int stmtCacheEntPerBucket = 8;		/* # entries/bucket		*/
-static stmtCacheEntry stmtCacheEntries[16384] = {{0, {0}, 0, 0, 0}};
+static stmtCacheEntry *stmtCacheEntries = NULL;
 
-static struct prepared_statement *find_prepared_statement(const char *name,
-				 struct connection * con, struct prepared_statement ** prev);
-static bool deallocate_one(int lineno, enum COMPAT_MODE c, struct connection * con,
-		 struct prepared_statement * prev, struct prepared_statement * this);
+static bool deallocate_one(int lineno, enum COMPAT_MODE c, struct connection *con,
+						   struct prepared_statement *prev, struct prepared_statement *this);
 
 static bool
 isvarchar(unsigned char c)
@@ -52,11 +52,56 @@ isvarchar(unsigned char c)
 	if (c >= 128)
 		return true;
 
-	return (false);
+	return false;
+}
+
+bool
+ecpg_register_prepared_stmt(struct statement *stmt)
+{
+	struct statement *prep_stmt;
+	struct prepared_statement *this;
+	struct connection *con = stmt->connection;
+	struct prepared_statement *prev = NULL;
+	int			lineno = stmt->lineno;
+
+	/* check if we already have prepared this statement */
+	this = ecpg_find_prepared_statement(stmt->name, con, &prev);
+	if (this && !deallocate_one(lineno, ECPG_COMPAT_PGSQL, con, prev, this))
+		return false;
+
+	/* allocate new statement */
+	this = (struct prepared_statement *) ecpg_alloc(sizeof(struct prepared_statement), lineno);
+	if (!this)
+		return false;
+
+	prep_stmt = (struct statement *) ecpg_alloc(sizeof(struct statement), lineno);
+	if (!prep_stmt)
+	{
+		ecpg_free(this);
+		return false;
+	}
+	memset(prep_stmt, 0, sizeof(struct statement));
+
+	/* create statement */
+	prep_stmt->lineno = lineno;
+	prep_stmt->connection = con;
+	prep_stmt->command = ecpg_strdup(stmt->command, lineno);
+	prep_stmt->inlist = prep_stmt->outlist = NULL;
+	this->name = ecpg_strdup(stmt->name, lineno);
+	this->stmt = prep_stmt;
+	this->prepared = true;
+
+	if (con->prep_stmts == NULL)
+		this->next = NULL;
+	else
+		this->next = con->prep_stmts;
+
+	con->prep_stmts = this;
+	return true;
 }
 
 static bool
-replace_variables(char **text, int lineno, bool questionmarks)
+replace_variables(char **text, int lineno)
 {
 	bool		string = false;
 	int			counter = 1,
@@ -74,9 +119,9 @@ replace_variables(char **text, int lineno, bool questionmarks)
 			ptr += 2;			/* skip  '::' */
 		else
 		{
+			/* a rough guess of the size we need: */
+			int			buffersize = sizeof(int) * CHAR_BIT * 10 / 3;
 			int			len;
-			int			buffersize = sizeof(int) * CHAR_BIT * 10 / 3;	/* a rough guess of the
-																		 * size we need */
 			char	   *buffer,
 					   *newcopy;
 
@@ -85,14 +130,15 @@ replace_variables(char **text, int lineno, bool questionmarks)
 
 			snprintf(buffer, buffersize, "$%d", counter++);
 
-			for (len = 1; (*text)[ptr + len] && isvarchar((*text)[ptr + len]); len++);
-			if (!(newcopy = (char *) ecpg_alloc(strlen(*text) -len + strlen(buffer) + 1, lineno)))
+			for (len = 1; (*text)[ptr + len] && isvarchar((*text)[ptr + len]); len++)
+				 /* skip */ ;
+			if (!(newcopy = (char *) ecpg_alloc(strlen(*text) - len + strlen(buffer) + 1, lineno)))
 			{
 				ecpg_free(buffer);
 				return false;
 			}
 
-			strncpy(newcopy, *text, ptr);
+			memcpy(newcopy, *text, ptr);
 			strcpy(newcopy + ptr, buffer);
 			strcat(newcopy, (*text) +ptr + len);
 
@@ -109,26 +155,13 @@ replace_variables(char **text, int lineno, bool questionmarks)
 	return true;
 }
 
-/* handle the EXEC SQL PREPARE statement */
-bool
-ECPGprepare(int lineno, const char *connection_name, const int questionmarks, const char *name, const char *variable)
+static bool
+prepare_common(int lineno, struct connection *con, const char *name, const char *variable)
 {
-	struct connection *con;
 	struct statement *stmt;
-	struct prepared_statement *this,
-			   *prev;
+	struct prepared_statement *this;
 	PGresult   *query;
 
-	con = ecpg_get_connection(connection_name);
-
-	if (!ecpg_init(con, connection_name, lineno))
-		return false;
-
-	/* check if we already have prepared this statement */
-	this = find_prepared_statement(name, con, &prev);
-	if (this && !deallocate_one(lineno, ECPG_COMPAT_PGSQL, con, prev, this))
-		return false;
-
 	/* allocate new statement */
 	this = (struct prepared_statement *) ecpg_alloc(sizeof(struct prepared_statement), lineno);
 	if (!this)
@@ -147,11 +180,11 @@ ECPGprepare(int lineno, const char *connection_name, const int questionmarks, co
 	stmt->command = ecpg_strdup(variable, lineno);
 	stmt->inlist = stmt->outlist = NULL;
 
-	/* if we have C variables in our statment replace them with '?' */
-	replace_variables(&(stmt->command), lineno, questionmarks);
+	/* if we have C variables in our statement replace them with '?' */
+	replace_variables(&(stmt->command), lineno);
 
 	/* add prepared statement to our list */
-	this->name = (char *) name;
+	this->name = ecpg_strdup(name, lineno);
 	this->stmt = stmt;
 
 	/* and finally really prepare the statement */
@@ -159,12 +192,13 @@ ECPGprepare(int lineno, const char *connection_name, const int questionmarks, co
 	if (!ecpg_check_PQresult(query, stmt->lineno, stmt->connection->connection, stmt->compat))
 	{
 		ecpg_free(stmt->command);
+		ecpg_free(this->name);
 		ecpg_free(this);
 		ecpg_free(stmt);
 		return false;
 	}
 
-	ecpg_log("ECPGprepare line %d: NAME: %s QUERY: %s\n", stmt->lineno, name, stmt->command);
+	ecpg_log("prepare_common on line %d: name %s; query: \"%s\"\n", stmt->lineno, name, stmt->command);
 	PQclear(query);
 	this->prepared = true;
 
@@ -177,14 +211,40 @@ ECPGprepare(int lineno, const char *connection_name, const int questionmarks, co
 	return true;
 }
 
-static struct prepared_statement *
-find_prepared_statement(const char *name,
-				 struct connection * con, struct prepared_statement ** prev_)
+/* handle the EXEC SQL PREPARE statement */
+/* questionmarks is not needed but remains in there for the time being to not change the API */
+bool
+ECPGprepare(int lineno, const char *connection_name, const bool questionmarks,
+			const char *name, const char *variable)
+{
+	struct connection *con;
+	struct prepared_statement *this,
+			   *prev;
+
+	(void) questionmarks;		/* quiet the compiler */
+
+	con = ecpg_get_connection(connection_name);
+	if (!ecpg_init(con, connection_name, lineno))
+		return false;
+
+	/* check if we already have prepared this statement */
+	this = ecpg_find_prepared_statement(name, con, &prev);
+	if (this && !deallocate_one(lineno, ECPG_COMPAT_PGSQL, con, prev, this))
+		return false;
+
+	return prepare_common(lineno, con, name, variable);
+}
+
+struct prepared_statement *
+ecpg_find_prepared_statement(const char *name,
+							 struct connection *con, struct prepared_statement **prev_)
 {
 	struct prepared_statement *this,
 			   *prev;
 
-	for (this = con->prep_stmts, prev = NULL; this != NULL; prev = this, this = this->next)
+	for (this = con->prep_stmts, prev = NULL;
+		 this != NULL;
+		 prev = this, this = this->next)
 	{
 		if (strcmp(this->name, name) == 0)
 		{
@@ -197,11 +257,12 @@ find_prepared_statement(const char *name,
 }
 
 static bool
-deallocate_one(int lineno, enum COMPAT_MODE c, struct connection * con, struct prepared_statement * prev, struct prepared_statement * this)
+deallocate_one(int lineno, enum COMPAT_MODE c, struct connection *con,
+			   struct prepared_statement *prev, struct prepared_statement *this)
 {
 	bool		r = false;
 
-	ecpg_log("ECPGdeallocate line %d: NAME: %s\n", lineno, this->name);
+	ecpg_log("deallocate_one on line %d: name %s\n", lineno, this->name);
 
 	/* first deallocate the statement in the backend */
 	if (this->prepared)
@@ -216,7 +277,9 @@ deallocate_one(int lineno, enum COMPAT_MODE c, struct connection * con, struct p
 			sprintf(text, "deallocate \"%s\"", this->name);
 			query = PQexec(this->stmt->connection->connection, text);
 			ecpg_free(text);
-			if (ecpg_check_PQresult(query, lineno, this->stmt->connection->connection, this->stmt->compat))
+			if (ecpg_check_PQresult(query, lineno,
+									this->stmt->connection->connection,
+									this->stmt->compat))
 			{
 				PQclear(query);
 				r = true;
@@ -237,6 +300,7 @@ deallocate_one(int lineno, enum COMPAT_MODE c, struct connection * con, struct p
 	/* okay, free all the resources */
 	ecpg_free(this->stmt->command);
 	ecpg_free(this->stmt);
+	ecpg_free(this->name);
 	if (prev != NULL)
 		prev->next = this->next;
 	else
@@ -255,11 +319,10 @@ ECPGdeallocate(int lineno, int c, const char *connection_name, const char *name)
 			   *prev;
 
 	con = ecpg_get_connection(connection_name);
-
 	if (!ecpg_init(con, connection_name, lineno))
 		return false;
 
-	this = find_prepared_statement(name, con, &prev);
+	this = ecpg_find_prepared_statement(name, con, &prev);
 	if (this)
 		return deallocate_one(lineno, c, con, prev, this);
 
@@ -271,7 +334,7 @@ ECPGdeallocate(int lineno, int c, const char *connection_name, const char *name)
 }
 
 bool
-ecpg_deallocate_all_conn(int lineno, enum COMPAT_MODE c, struct connection * con)
+ecpg_deallocate_all_conn(int lineno, enum COMPAT_MODE c, struct connection *con)
 {
 	/* deallocate all prepared statements */
 	while (con->prep_stmts)
@@ -286,23 +349,27 @@ ecpg_deallocate_all_conn(int lineno, enum COMPAT_MODE c, struct connection * con
 bool
 ECPGdeallocate_all(int lineno, int compat, const char *connection_name)
 {
-	return ecpg_deallocate_all_conn(lineno, compat, ecpg_get_connection(connection_name));
+	return ecpg_deallocate_all_conn(lineno, compat,
+									ecpg_get_connection(connection_name));
 }
 
 char *
-ecpg_prepared(const char *name, struct connection * con, int lineno)
+ecpg_prepared(const char *name, struct connection *con)
 {
 	struct prepared_statement *this;
 
-	this = find_prepared_statement(name, con, NULL);
+	this = ecpg_find_prepared_statement(name, con, NULL);
 	return this ? this->stmt->command : NULL;
 }
 
 /* return the prepared statement */
+/* lineno is not used here, but kept in to not break API */
 char *
 ECPGprepared_statement(const char *connection_name, const char *name, int lineno)
 {
-	return ecpg_prepared(name, ecpg_get_connection(connection_name), lineno);
+	(void) lineno;				/* keep the compiler quiet */
+
+	return ecpg_prepared(name, ecpg_get_connection(connection_name));
 }
 
 /*
@@ -315,27 +382,28 @@ HashStmt(const char *ecpgQuery)
 				bucketNo,
 				hashLeng,
 				stmtLeng;
-	long long	hashVal,
+	uint64		hashVal,
 				rotVal;
 
 	stmtLeng = strlen(ecpgQuery);
-	hashLeng = 50;				/* use 1st 50 characters of statement		*/
-	if (hashLeng > stmtLeng)	/* if the statement isn't that long         */
-		hashLeng = stmtLeng;	/* use its actual length			   */
+	hashLeng = 50;				/* use 1st 50 characters of statement */
+	if (hashLeng > stmtLeng)	/* if the statement isn't that long */
+		hashLeng = stmtLeng;	/* use its actual length */
 
 	hashVal = 0;
 	for (stmtIx = 0; stmtIx < hashLeng; ++stmtIx)
 	{
-		hashVal = hashVal + (int) ecpgQuery[stmtIx];
+		hashVal = hashVal + (unsigned char) ecpgQuery[stmtIx];
+		/* rotate 32-bit hash value left 13 bits */
 		hashVal = hashVal << 13;
-		rotVal = (hashVal & 0x1fff00000000LL) >> 32;
-		hashVal = (hashVal & 0xffffffffLL) | rotVal;
+		rotVal = (hashVal & UINT64CONST(0x1fff00000000)) >> 32;
+		hashVal = (hashVal & UINT64CONST(0xffffffff)) | rotVal;
 	}
 
 	bucketNo = hashVal % stmtCacheNBuckets;
-	bucketNo += 1;				/* don't use bucket # 0         */
 
-	return (bucketNo * stmtCacheEntPerBucket);
+	/* Add 1 so that array entry 0 is never used */
+	return bucketNo * stmtCacheEntPerBucket + 1;
 }
 
 /*
@@ -349,25 +417,29 @@ SearchStmtCache(const char *ecpgQuery)
 	int			entNo,
 				entIx;
 
-/* hash the statement			*/
+	/* quick failure if cache not set up */
+	if (stmtCacheEntries == NULL)
+		return 0;
+
+	/* hash the statement */
 	entNo = HashStmt(ecpgQuery);
 
-/* search the cache		*/
+	/* search the cache */
 	for (entIx = 0; entIx < stmtCacheEntPerBucket; ++entIx)
 	{
-		if (stmtCacheEntries[entNo].stmtID[0])	/* check if entry is in use		*/
+		if (stmtCacheEntries[entNo].stmtID[0])	/* check if entry is in use */
 		{
-			if (!strcmp(ecpgQuery, stmtCacheEntries[entNo].ecpgQuery))
-				break;			/* found it		*/
+			if (strcmp(ecpgQuery, stmtCacheEntries[entNo].ecpgQuery) == 0)
+				break;			/* found it */
 		}
-		++entNo;				/* incr entry #		*/
+		++entNo;				/* incr entry # */
 	}
 
-/* if entry wasn't found - set entry # to zero  */
+	/* if entry wasn't found - set entry # to zero */
 	if (entIx >= stmtCacheEntPerBucket)
 		entNo = 0;
 
-	return (entNo);
+	return entNo;
 }
 
 /*
@@ -376,33 +448,39 @@ SearchStmtCache(const char *ecpgQuery)
  *	 OR  negative error code
  */
 static int
-ecpg_freeStmtCacheEntry(int lineno, int compat, int entNo)		/* entry # to free */
+ecpg_freeStmtCacheEntry(int lineno, int compat,
+						int entNo)	/* entry # to free */
 {
 	stmtCacheEntry *entry;
 	struct connection *con;
- 	struct prepared_statement *this, *prev;
+	struct prepared_statement *this,
+			   *prev;
+
+	/* fail if cache isn't set up */
+	if (stmtCacheEntries == NULL)
+		return -1;
 
 	entry = &stmtCacheEntries[entNo];
-	if (!entry->stmtID[0])		/* return if the entry isn't in use     */
-		return (0);
+	if (!entry->stmtID[0])		/* return if the entry isn't in use */
+		return 0;
 
 	con = ecpg_get_connection(entry->connection);
 
-	/* free the 'prepared_statement' list entry       */
-	this = find_prepared_statement(entry->stmtID, con, &prev);
+	/* free the 'prepared_statement' list entry */
+	this = ecpg_find_prepared_statement(entry->stmtID, con, &prev);
 	if (this && !deallocate_one(lineno, compat, con, prev, this))
-		return (-1);
+		return -1;
 
 	entry->stmtID[0] = '\0';
 
-	/* free the memory used by the cache entry		*/
+	/* free the memory used by the cache entry */
 	if (entry->ecpgQuery)
 	{
 		ecpg_free(entry->ecpgQuery);
 		entry->ecpgQuery = 0;
 	}
 
-	return (entNo);
+	return entNo;
 }
 
 /*
@@ -410,11 +488,11 @@ ecpg_freeStmtCacheEntry(int lineno, int compat, int entNo)		/* entry # to free *
  * returns entry # in cache used  OR  negative error code
  */
 static int
-AddStmtToCache(int lineno,		/* line # of statement		*/
-			   char *stmtID,	/* statement ID				*/
-			   const char *connection,	/* connection				*/
-			   int compat, 			/* compatibility level */
-			   const char *ecpgQuery)	/* query					*/
+AddStmtToCache(int lineno,		/* line # of statement */
+			   const char *stmtID,	/* statement ID */
+			   const char *connection,	/* connection */
+			   int compat,		/* compatibility level */
+			   const char *ecpgQuery)	/* query */
 {
 	int			ix,
 				initEntNo,
@@ -422,73 +500,103 @@ AddStmtToCache(int lineno,		/* line # of statement		*/
 				entNo;
 	stmtCacheEntry *entry;
 
-/* hash the statement																	*/
+	/* allocate and zero cache array if we haven't already */
+	if (stmtCacheEntries == NULL)
+	{
+		stmtCacheEntries = (stmtCacheEntry *)
+			ecpg_alloc(sizeof(stmtCacheEntry) * stmtCacheArraySize, lineno);
+		if (stmtCacheEntries == NULL)
+			return -1;
+	}
+
+	/* hash the statement */
 	initEntNo = HashStmt(ecpgQuery);
 
-/* search for an unused entry															*/
+	/* search for an unused entry */
 	entNo = initEntNo;			/* start with the initial entry # for the
-								 * bucket	 */
-	luEntNo = initEntNo;		/* use it as the initial 'least used' entry			*/
+								 * bucket */
+	luEntNo = initEntNo;		/* use it as the initial 'least used' entry */
 	for (ix = 0; ix < stmtCacheEntPerBucket; ++ix)
 	{
 		entry = &stmtCacheEntries[entNo];
-		if (!entry->stmtID[0])	/* unused entry  -	use it			*/
+		if (!entry->stmtID[0])	/* unused entry  -	use it */
 			break;
 		if (entry->execs < stmtCacheEntries[luEntNo].execs)
-			luEntNo = entNo;	/* save new 'least used' entry		*/
-		++entNo;				/* increment entry #				*/
+			luEntNo = entNo;	/* save new 'least used' entry */
+		++entNo;				/* increment entry # */
 	}
 
-/* if no unused entries were found - use the 'least used' entry found in the bucket		*/
-	if (ix >= stmtCacheEntPerBucket)	/* if no unused entries were found	*/
-		entNo = luEntNo;		/* re-use the 'least used' entry	*/
+	/*
+	 * if no unused entries were found, re-use the 'least used' entry found in
+	 * the bucket
+	 */
+	if (ix >= stmtCacheEntPerBucket)
+		entNo = luEntNo;
 
-/* 'entNo' is the entry to use - make sure its free										*/
+	/* 'entNo' is the entry to use - make sure its free */
 	if (ecpg_freeStmtCacheEntry(lineno, compat, entNo) < 0)
-		return (-1);
+		return -1;
 
-/* add the query to the entry															*/
+	/* add the query to the entry */
 	entry = &stmtCacheEntries[entNo];
 	entry->lineno = lineno;
 	entry->ecpgQuery = ecpg_strdup(ecpgQuery, lineno);
-	entry->connection = (char *) connection;
+	entry->connection = connection;
 	entry->execs = 0;
 	memcpy(entry->stmtID, stmtID, sizeof(entry->stmtID));
 
-	return (entNo);
+	return entNo;
 }
 
-/* handle cache and preparation of statments in auto-prepare mode */
+/* handle cache and preparation of statements in auto-prepare mode */
 bool
-ecpg_auto_prepare(int lineno, const char *connection_name, int compat, const int questionmarks, char **name, const char *query)
+ecpg_auto_prepare(int lineno, const char *connection_name, const int compat, char **name, const char *query)
 {
 	int			entNo;
 
-	/* search the statement cache for this statement	*/
+	/* search the statement cache for this statement */
 	entNo = SearchStmtCache(query);
 
-	/* if not found - add the statement to the cache	*/
+	/* if not found - add the statement to the cache */
 	if (entNo)
 	{
-		ecpg_log("ecpg_auto_prepare line %d: stmt found in cache, entry %d\n", lineno, entNo);
-		*name = ecpg_strdup(stmtCacheEntries[entNo].stmtID, lineno);
+		char	   *stmtID;
+		struct connection *con;
+		struct prepared_statement *prep;
+
+		ecpg_log("ecpg_auto_prepare on line %d: statement found in cache; entry %d\n", lineno, entNo);
+
+		stmtID = stmtCacheEntries[entNo].stmtID;
+
+		con = ecpg_get_connection(connection_name);
+		prep = ecpg_find_prepared_statement(stmtID, con, NULL);
+		/* This prepared name doesn't exist on this connection. */
+		if (!prep && !prepare_common(lineno, con, stmtID, query))
+			return false;
+
+		*name = ecpg_strdup(stmtID, lineno);
 	}
 	else
 	{
-		ecpg_log("ecpg_auto_prepare line %d: stmt not in cache; inserting\n", lineno);
+		char		stmtID[STMTID_SIZE];
+
+		ecpg_log("ecpg_auto_prepare on line %d: statement not in cache; inserting\n", lineno);
 
 		/* generate a statement ID */
-		*name = (char *) ecpg_alloc(STMTID_SIZE, lineno);
-		sprintf(*name, "ecpg%d", nextStmtID++);
+		sprintf(stmtID, "ecpg%d", nextStmtID++);
 
-		if (!ECPGprepare(lineno, connection_name, questionmarks, ecpg_strdup(*name, lineno), query))
-			return (false);
-		if (AddStmtToCache(lineno, *name, connection_name, compat, query) < 0)
-			return (false);
+		if (!ECPGprepare(lineno, connection_name, 0, stmtID, query))
+			return false;
+
+		entNo = AddStmtToCache(lineno, stmtID, connection_name, compat, query);
+		if (entNo < 0)
+			return false;
+
+		*name = ecpg_strdup(stmtID, lineno);
 	}
 
 	/* increase usage counter */
 	stmtCacheEntries[entNo].execs++;
 
-	return (true);
+	return true;
 }
